 /************************************************************************
 Author: Umit Ogras  (uogras@ece.cmu.edu)
 File name: add_rand_link_rtable.cpp
 Date Written:  <Sat, Jan 8 13:08:25, 2005>
 Last Modified: <Thu, Jan 27 11:03, 2005>
 Description: Function that application specific long-range links to 
 a mesh network and generates the routing table.
 Modification: It does not write bidirectional links anymore. Writing only 
 one direction is sufficient to generate two uni-directional links between 
 the nodes specified in the file.
************************************************************************/
#include <iostream>
#include <fstream>
#include <cstdio>
#include <vector>

using namespace std;

int gen_arch_config(int rows,int cols,int count,vector<int> long_links,char *arch_file) {
  
  int *nodes = new int[rows*cols];
  for (int i=0; i<rows*cols; i++) 
    nodes[i]=-1;                                   // No extra link is assigned to the nodes yet
 
  ofstream arch_config(arch_file);                 // Open the architecture file to write to
  if (!arch_config.is_open() || !arch_config.good()) {
    cerr <<"Traffic file cannot be opened \n";
    return -1;
  }
 
  for (int i=0; i<count; i+=2) {
    arch_config <<"@NODE \t" <<long_links[i] <<endl;
    arch_config <<"random_link " << long_links[i+1] <<endl;
    nodes[long_links[i]] = long_links[i+1];
    nodes[long_links[i+1]] = long_links[i];
  }
  
  unsigned int dir = 0;
  unsigned int rlink = 0;
  int rlink_x=0, rlink_y=0, rdir=-1;
  bool flag=true;
  for (int src_x=0; src_x<cols; src_x++)                        // From each router
    for (int src_y=0; src_y<rows; src_y++) {                    // (src_x,src_y): source row and source columns
      arch_config <<"@NODE \t" <<src_y+src_x*rows <<endl;
      arch_config <<"routing_scheme  table ";
      for (int dst_x=0; dst_x<cols; dst_x++)                       // To each router 
	for (int dst_y=0; dst_y<rows; dst_y++) {                   // (dst_x,dst_y): destination row and column
          /* printf("(%d,%d)=%d --> (%d,%d)=%d\n",
             src_x,src_y,src_y+src_x*rows,dst_x,dst_y,dst_y+dst_x*rows); */
	  if ( (src_x==dst_x) && (src_y==dst_y) )
	    dir = 0;
	  else if (src_x < dst_x)
	    dir = 3;
	  else if (src_x > dst_x)
	    dir = 4;
	  else if (src_y < dst_y)
	    dir = 1;
	  else if (src_y > dst_y)
	    dir = 2;                                   // Until here find XY decision
          
	  if ( (dir>0) && (nodes[src_y+src_x*rows] >= 0) ) { // There is a random link assigned
	    rlink = nodes[src_y+src_x*rows];                 // The random link connects the source to rlink
	    rlink_y = rlink % rows;
	    rlink_x = int (rlink/rows);                     // Now find the direction Random Link drives the source
            
	    if (rlink_x > src_x)                               // Random link drives to pos X
	      if (rlink_y > src_y)                             // Random link drives to pos Y
		rdir = 7;                              // NorthEast: includes X_to_Y turn
	      else if (rlink_y ==  src_y)
		rdir = 6 ;                             // East: includes *_to_X turn
	      else 
		rdir = 5;                              // SouthEast
	    else if (rlink_x < src_x)                          // Random link drives to neg X
	      if (rlink_y > src_y)                             // Random link drives to pos Y
		rdir = 1;                              // NorthWest
	      else if (rlink_y ==  src_y)
		rdir = 2 ;                             // West
	      else 
		rdir = 3;                              // SouthWest
	    else
	      if (rlink_y > src_y)                             // Random link drives to pos Y
		rdir = 0;                              // North
	      else if (rlink_y <  src_y)
		rdir = 4 ;                             // South
	      else {	
		cout <<"Original direction is: "<<dir <<" \n";
		cout <<rlink_x <<":rlink_x, " <<src_x <<":src_x," <<rlink_y <<":rlink_y, " <<src_y <<":src_y \n";
                cerr <<"Failed to find the requested output direction!\n";
		return -1;
	      }
	    
	    if ( (rdir==3) || (rdir==4) || (rdir==5) ) 
	      if ( (dir==3)||(dir==4) || ((dir==2)&&(rlink_x!=dst_x)) )    // Turn negY_to_X is not allowed
		flag = false;
	    
	    if ( (rlink_x==dst_x) && (rlink_y==dst_y) )                   // Random link is connected to the destination
	      flag = true;                              // Overwrite the previous value
	    
	    if ( ((abs(rlink_y-dst_y)+abs(rlink_x-dst_x)) < ((abs(src_x-dst_x)+abs(src_y-dst_y)-1))) && flag ) 
	      dir =5;                                   // Random link provides a shorter path
            
	    flag = true;
	  }
	  arch_config << dir <<" ";
	} // for (int dst_y=0...
      arch_config <<endl;
    }  
}
